#include "NetSim.h"
#include <data/metadata.h>
#include <data/filters.h>
#include <data/xmipp_filename.h>
#include <sstream>
#include <time.h>
#include <mpi.h>

/*
this program compute adjacency matrix from raw particle image set
using xmipp's api to handle image IO and using MPI to accrelate
*/

using namespace std;
using namespace MPI;

Comm &comm = COMM_WORLD;
int id, numprocs;

void readImage(MetaData &md, Image<double> &I, size_t objId, bool applyGeo)
{
    if (applyGeo)
        I.readApplyGeo(md, objId);
    else
    {
        FileName fnImg;
        md.getValue(MDL_IMAGE, fnImg, objId);
        I.read(fnImg);
    }
    I().setXmippOrigin();
    I().statisticsAdjust(0, 1);
}

int main(int argc, char **argv)
{
    time_t start, end;
    start=time(NULL);
    Init(argc, argv);
    id=comm.Get_rank();
    numprocs=comm.Get_size();

    //get image information
	MetaData md;
	md.read(argv[1]);
    stringstream sK(argv[2]);
    int k;
    sK>>k;
    FileName fn(argv[1]);

	int Nimgs = md.size();
    size_t Xdim, Ydim, Zdim, Ndim;
    double sigma = 0;
    getImageSize(md, Xdim, Ydim, Zdim, Ndim);

    //compute sigma for correntorpy
    cout<<"Computing sigma..."<<endl;
    MultidimArray<int> mask(Ydim,Xdim);
    mask.setXmippOrigin();
    BinaryCircularMask(mask, Xdim/2, INNER_MASK);

    for(int i=1; i<=Nimgs; i++){
        Image<double> I;
        readImage(md, I, i, false);
        double avg, stddev;
        I().computeAvgStdev_within_binary_mask(mask, avg, stddev);
        sigma += stddev;
    }
    sigma *= sqrt(2.0) / Nimgs;

    //compute correntorpy matrix, using MPI
	cout<<"Computing corrM..."<<endl;

    MultidimArray<double> corrM(Nimgs,Nimgs);
    MultidimArray<double> adjancent(Nimgs, Nimgs);

    for(int i=0; i<Nimgs; i++){
        if(i%numprocs!=id)
            continue;
        Image<double> img1;
        readImage(md, img1, i+1, false);
        for(int j=0; j<Nimgs; j++){
            if(i==j){
                dAij(corrM, i, j)=1;
                continue;
            }
            if(j<i){
                continue;
            }
            Image<double> img2;
            readImage(md, img2, j+1, false);
            dAij(corrM, i, j) = correntropy(img1.data, img2.data, sqrt(2)*sigma);
        }
    }
    comm.Barrier();
    for(int i=0; i<Nimgs; i++){
        int source = i%numprocs;
        comm.Bcast(&dAij(corrM, i, i), corrM.xdim - i, MPI_DOUBLE, source); 
    }
    comm.Barrier();

    if(id!=0){
        Finalize();
        return 0;
    }
    for(int i=0; i<Nimgs; i++)
        for(int j=0; j<i; j++) dAij(corrM, i, j) = dAij(corrM, j, i);

    //create mutual knn network based on correntorpy matrix
    Image<double> out=corrM;out.write(fn.getDir().getString() + "/corr.stk");
    CreateAdjacentMatrix(corrM, adjancent, k, Cmp(false));
    Image<double> aout = adjancent; aout.write(fn.getDir().getString() + "/adj.stk");
    cout<<"Graph Constructed"<<endl;

    Finalize();
    end=time(NULL);
    cout<<"all process cost "<<difftime(end, start)<<" seconds"<<endl;
    
    return 0;
}
